import { StyledProps } from 'styled-components';
import { ColorNames } from './colors';
import { Spacing } from './spacing';
import { ShadowNames, shadowsToString } from './shadows';

/**
 * Returns a typography from the theme.
 */
export function typography<T = any>(
  size: PropPicker<string>,
  weight: PropPicker<string>,
  variant: PropPicker<string>
) {
  return function(props: StyledProps<T>) {
    const pickSize = pickProp(size);
    const pickWeight = pickProp(weight);
    const pickVariant = pickProp(variant);

    const sizeConfig = props.theme.typographies[pickSize(props)];
    const fontWeightConfig = sizeConfig.weights[pickWeight(props)];
    const variantConfig = fontWeightConfig.variants[pickVariant(props)];

    return `
      font-size: ${sizeConfig.fontSize};
      line-height: ${sizeConfig.lineHeight};
      letter-spacing: ${sizeConfig.letterSpacing};
      font-weight: ${fontWeightConfig.fontWeight};
      color: ${variantConfig};
    `;
  };
}

/**
 * Returns box shadow from the theme.
 */
export function boxShadow<T = any>(name: PropPicker<ShadowNames>) {
  return function(props: StyledProps<T>) {
    const pickShadow = pickProp(name);
    const shadows = props.theme.shadows[pickShadow(props)];
    return `box-shadow: ${shadowsToString(shadows)};`;
  };
}

/**
 * Returns a space from the theme.
 */
export function spacing<T = any>(space: PropPicker<Spacing>) {
  return function(props: StyledProps<T>) {
    const pickSpace = pickProp(space);
    return props.theme.spacing[pickSpace(props)];
  };
}

/**
 * Returns a color from the theme.
 */
export function color<T = any>(
  name: PropPicker<ColorNames>,
  tone: PropPicker<string>
) {
  return function(props: StyledProps<T>) {
    const pickName = pickProp(name);
    const pickTone = pickProp(tone);
    return props.theme.palette[pickName(props)][pickTone(props)];
  };
}

export type PropPicker<T = any> = T | ((props: any) => T);

/**
 * Returns a curry function that returns the value of the `attr` from `props`.
 *
 *    const props = { tone: 50 }
 *    const tone = fromProp('tone');
 *    tone(props) == 50
 */
export function fromProp<T = any>(attr: string) {
  return function(props: any): T {
    return props[attr];
  };
}

export function pickProp<T = any>(pickerOrValue: PropPicker<T>) {
  return function(props: any): T {
    return pickerOrValue instanceof Function
      ? pickerOrValue(props)
      : pickerOrValue;
  };
}

/**
 * Returns styles based on the `attr` value from the props.
 * @param attr the attributes selector of the props.
 * @param branches defines the styles matching the key of the object to the
 * possible values of `attr` from props.
 */
export function styledBy<R = any, T = any>(
  attr: (props: any) => T | T,
  branches: Record<any, any>
) {
  return function(props: any): R {
    const value = attr instanceof Function ? attr(props) : props[attr];

    const branch = branches[value];
    const pickBranch = pickProp(branch);

    return pickBranch(props);
  };
}
